// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"bytes"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"strings"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
	yaml "gopkg.in/yaml.v2"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
)

// Indent a block of text
func Indent(data, indent []byte) []byte {
	b := bytes.Buffer{}
	w := IndentWriter{indent, &b, false}
	_, _ = w.Write(data)
	return b.Bytes()
}

// IndentString a block of text
func IndentString(data, indent string) string {
	b := bytes.Buffer{}
	w := IndentWriter{[]byte(indent), &b, false}
	_, _ = w.Write([]byte(data))
	return b.String()
}

// IndentWriter writes with added indentation
type IndentWriter struct {
	indent      []byte
	out         io.Writer
	linestarted bool
}

func (w *IndentWriter) Write(data []byte) (int, error) {
	lines := bytes.Split(data, []byte("\n"))
	if !w.linestarted {
		if _, err := w.out.Write(w.indent); err != nil {
			return 0, err
		}
	}
	if n, err := w.out.Write(append(lines[0], '\n')); err != nil {
		return n, err
	}
	total := len(lines[0])
	for _, line := range lines[1 : len(lines)-1] {
		if _, err := w.out.Write(w.indent); err != nil {
			return total, err
		}
		n, err := w.out.Write(append(line, '\n'))
		total += n
		if err != nil {
			return total, err
		}
	}
	lastline := lines[len(lines)-1]
	if len(lastline) == 0 {
		w.linestarted = false
	} else {
		w.linestarted = true
		if _, err := w.out.Write(w.indent); err != nil {
			return total, err
		}
		n, err := w.out.Write(lastline)
		total += n
		if err != nil {
			return total, err
		}
	}
	return total, nil
}

func pipelineInfo(identifier, version string) api.PipelineInfo {
	var info api.PipelineInfo
	if id, err := api.UUIDFromString(identifier); err == nil {
		info.ID = id
	} else {
		info.Name = identifier
		info.Version = version
	}
	return info
}

func pipelineList(cmd *cobra.Command, args []string) {
	c, err := GetClient(false)
	if err != nil {
		return
	}
	if len(args) > 1 {
		usage(cmd)
	}

	extended := cmd.Flag("extended").Value.String() == "true"

	filter, err := parsePipelineFilterFlags(cmd)
	if err != nil {
		Logger.Errorf("%s", err)
		return
	}

	pipelines, graphs, err := filterPipelineList(c, filter)
	if err != nil {
		Logger.Errorf("%s", err)
		return
	}

	table := tablewriter.NewWriter(os.Stdout)
	headers := []string{"id", "pipeline", "version", "status"}
	if extended {
		headers = append(
			headers, "type", "sourcematch", "emitter(s)", "workers(s)", "consumer(s)")
	}
	table.SetHeader(headers)
	if !extended {
		table.SetBorder(false)
		table.SetHeaderLine(false)
		table.SetColumnSeparator(" ")
	} else {
		table.SetRowLine(true)
	}
	table.SetAlignment(tablewriter.ALIGN_LEFT)

	for i, info := range pipelines {
		var graph *storage.Graph
		if extended {
			if graphs != nil {
				graph = graphs[i]
			}
			if graph == nil {
				b := bytes.NewBuffer(nil)
				if err := pipelineGet(c, info.Name, info.Version, b); err != nil {
					Logger.Errorf("%s", err)
					return
				}
				graph = &storage.Graph{}
				if err := yaml.Unmarshal(b.Bytes(), graph); err != nil {
					Logger.Errorf("%s", err)
					return
				}
			}
		}
		version := info.Version
		if version == "" {
			version = "<none>"
		}
		values := []string{
			info.ID.String(),
			info.Name,
			version,
			info.Status.String(),
		}
		if extended {
			var (
				sourcematch, emitters, workers, consumers []string
				requestreply                              bool
			)
			for _, node := range graph.Nodes {
				if node.Type == storage.RequestorNode {
					requestreply = true
				}
				switch node.Type {
				case storage.EmitterNode, storage.RequestorNode:
					if node.SourceMatch != nil {
						sourcematch = append(sourcematch, node.SourceMatch.EventTypes...)
					}
					for _, id := range node.ActorIDs {
						emitters = append(emitters, id.String())
					}
					emitters = append(emitters, node.Actors...)
					for _, role := range node.Roles {
						emitters = append(emitters, "role:"+role)
					}
				case storage.WorkerNode:
					for _, id := range node.ActorIDs {
						workers = append(workers, id.String())
					}
					workers = append(workers, node.Actors...)
					for _, role := range node.Roles {
						workers = append(workers, "role:"+role)
					}
				case storage.ConsumerNode:
					for _, id := range node.ActorIDs {
						consumers = append(consumers, id.String())
					}
					consumers = append(consumers, node.Actors...)
					for _, role := range node.Roles {
						consumers = append(consumers, "role:"+role)
					}
				}
			}
			if requestreply {
				values = append(values, "request/reply")
			} else {
				values = append(values, "one way")
			}
			values = append(values,
				strings.Join(sourcematch, "\n"),
				strings.Join(emitters, "\n"),
				strings.Join(workers, "\n"),
				strings.Join(consumers, "\n"))
		}
		table.Append(values)
	}
	table.Render()
}

func pipelineGet(c *xbus.Client, identifier, version string, out io.Writer) error {
	info := pipelineInfo(identifier, version)

	str, err := c.PipelineGet(info)
	if err != nil {
		return err
	}

	var graph storage.Graph
	if err := yaml.Unmarshal([]byte(str), &graph); err != nil {
		return err
	}
	sortNodeSlice(graph.Nodes)
	sortEdgeSlice(graph.Edges)
	b, err := yaml.Marshal(graph)
	if err != nil {
		return err
	}

	if _, err := out.Write(b); err != nil {
		return err
	}
	return nil
}

func pipelineSave(identifier, version, filename string) error {
	c, err := GetClient(false)
	if err != nil {
		return err
	}

	content, err := ioutil.ReadFile(filename)
	if err != nil {
		return fmt.Errorf("cannot read pipeline definition file: %w", err)
	}

	info := pipelineInfo(identifier, version)

	info, _, validationMessages, err := c.PipelineSave(info, string(content))
	if len(validationMessages) != 0 {
		fmt.Println("Validation messages:")
		for _, msg := range validationMessages {
			fmt.Println("-", msg)
		}
	}

	if err != nil {
		return err
	}

	Logger.Noticef("Successfully saved pipeline %s-%s graph", info.Name, info.Version)
	return nil
}

func pipelineStatus(identifier, version string, status api.PipelineInfo_Status) error {
	c, err := GetClient(false)
	if err != nil {
		return err
	}

	info := pipelineInfo(identifier, version)
	info.Status = status

	info, err = c.PipelineSetStatus(info)
	if err != nil {
		return err
	}
	Logger.Noticef("%s-%s status changed to %s", info.Name, info.Version, info.Status)
	return nil
}

var pipelineCmd = &cobra.Command{
	Use:   "pipeline",
	Short: "Pipeline configuration subcommands",
	Long:  "Pipeline subcommands allow to list and modify pipelines",
}

var pipelineListCmd = &cobra.Command{
	Use:   "list [name]",
	Short: "List pipelines, optionnaly filtered by name",
	Run:   pipelineList,
}

var pipelineGetCmd = &cobra.Command{
	Use:   "get <name or id> [version]",
	Short: "Get a pipeline definition",
	Long:  "Get a pipeline definition (its graph) and write it to stdout",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) == 0 || len(args) > 2 {
			usage(cmd)
		}
		c, err := GetClient(false)
		if err != nil {
			Logger.Errorf("%s", err)
			os.Exit(-1)
		}

		var (
			identifier = args[0]
			version    string
			filename   = cmd.Flag("output").Value.String()
		)
		if len(args) == 2 {
			version = args[1]
		}
		var out io.Writer
		if filename == "-" {
			out = os.Stdout
		} else {
			f, err := os.Create(filename)
			if err != nil {
				Logger.Errorf("Error opening output file: %s", err)
				os.Exit(-1)
			}
			defer func() {
				if err := f.Close(); err != nil {
					Logger.Errorf("Error saving pipeline: %s", err)
					os.Exit(-1)
				}
			}()
			out = f
		}
		if err := pipelineGet(c, identifier, version, out); err != nil {
			Logger.Errorf("Error getting pipeline: %s", err)
			os.Exit(-1)
		}
	},
}

var pipelineSaveCmd = &cobra.Command{
	Use:   "save <name or id> [version] <filepath>",
	Short: "Save a pipeline",
	Long:  "Save a pipeline definition (its graph)",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) < 2 || len(args) > 3 {
			usage(cmd)
		}
		var (
			identifier = args[0]
			version    string
			filename   string
		)
		if len(args) == 2 {
			filename = args[1]
		} else {
			version = args[1]
			filename = args[2]
		}
		if err := pipelineSave(identifier, version, filename); err != nil {
			Logger.Errorf("Error saving pipeline: %s", err)
			os.Exit(-1)
		}
	},
}

var pipelineStatusCmd = &cobra.Command{
	Use:   "status <name or id> [version] [--active|--inactive]",
	Short: "Change the status of a pipeline version",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) < 1 || len(args) > 2 {
			usage(cmd)
		}
		var (
			identifier = args[0]
			version    string
		)
		if len(args) == 2 {
			version = args[1]
		}
		var status api.PipelineInfo_Status
		if cmd.Flag("active").Value.String() == "true" {
			status = api.PipelineInfo_ACTIVE
		} else if cmd.Flag("inactive").Value.String() == "true" {
			status = api.PipelineInfo_INACTIVE
		}
		if err := pipelineStatus(identifier, version, status); err != nil {
			Logger.Errorf("Error setting pipeline status: %s", err)
			os.Exit(-1)
		}
	},
}

func init() {
	RootCmd.AddCommand(pipelineCmd)
	pipelineListCmd.Flags().Bool("extended", false, "Print more information on the pipelines")
	addPipelineFilterFlags(pipelineListCmd.Flags(), false)
	pipelineCmd.AddCommand(pipelineListCmd)
	pipelineCmd.AddCommand(pipelineGetCmd)
	pipelineCmd.AddCommand(pipelineSaveCmd)
	pipelineCmd.AddCommand(pipelineStatusCmd)

	pipelineStatusCmd.Flags().Bool("active", false, "Activate the pipeline")
	pipelineStatusCmd.Flags().Bool("inactive", false, "Deactivate the pipeline")

	pipelineGetCmd.Flags().StringP("output", "o", "-", "Target filename (default to stdout)")
}
