package pipeline

import (
	"fmt"

	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
)

// setEqual compare two slices of strings and return true if they contain
// the same elements no matter the order
// The sets must not contain twice the same element
func stringSetEqual(s1, s2 []string) bool {
	if len(s1) != len(s2) {
		return false
	}
	for _, v1 := range s1 {
		found := false
		for _, v2 := range s2 {
			if v1 == v2 {
				found = true
				break
			}
		}
		if !found {
			return false
		}
	}
	return true
}

// stringSetContains returns true if the string list contains the given value
func stringSetContains(set []string, s string) bool {
	for _, v := range set {
		if s == v {
			return true
		}
	}
	return false
}

// uuidSetContains returns true if the uuid list contains the given value
func uuidSetContains(set []api.UUID, s api.UUID) bool {
	for _, v := range set {
		if s == v {
			return true
		}
	}
	return false
}

// stringSetOverlap returns true if the string sets have common elements
func stringSetOverlap(s1, s2 []string) bool {
	for _, v := range s1 {
		if stringSetContains(s2, v) {
			return true
		}
	}
	return false
}

// NodeType is the type of Node
type NodeType int

const (
	// EmitterNode is the type of emitter nodes
	EmitterNode NodeType = iota
	// ConsumerNode is the type of consumer nodes
	ConsumerNode
	// WorkerNode is the type of worker nodes
	WorkerNode
)

type Match struct {
	Pipeline *Pipeline
	Output   storage.NodeOutputRef
	Priority storage.MatchPriority
}

// NewPipeline creates a graph
func NewPipeline(pipeline storage.Pipeline) *Pipeline {
	return &Pipeline{
		Pipeline: pipeline,
	}
}

// Pipeline is a single xbus pipeline
type Pipeline struct {
	storage.Pipeline `yaml:",inline"`
}

// GetDestinations returns the inputs on which to send envelope coming from given outputs
func (p Pipeline) GetDestinations(output storage.NodeOutputRef) ([]storage.NodeInputRef, error) {
	var dests []storage.NodeInputRef
	for _, edge := range p.Graph.Edges {
		if output == edge.From {
			dests = append(dests, edge.To)
		}
	}
	return dests, nil
}

// Match returns the source input corresponding to the emitter & eventtypes
// and true if it exists, empty value and false if not
func (p *Pipeline) Match(
	emitter api.Actor, output string, eventtypes []string,
) *Match {
	for _, node := range p.Graph.Nodes {
		if (node.Type == storage.EmitterNode && output == "default" ||
			node.Type == storage.RequestorNode && output == "request") &&
			(stringSetEqual(node.SourceMatch.EventTypes, eventtypes) ||
				node.SourceMatch.ByActor) &&
			(len(node.ActorIDs) == 0 || uuidSetContains(node.ActorIDs, emitter.ID)) &&
			(len(node.Actors) == 0 || stringSetContains(node.Actors, emitter.Name)) &&
			(len(node.Roles) == 0 || stringSetOverlap(node.Roles, emitter.Roles)) {
			var priority storage.MatchPriority

			switch {
			case node.SourceMatch.Priority != storage.MatchPriorityNone:
				priority = node.SourceMatch.Priority
			case !node.SourceMatch.ByActor && (len(node.ActorIDs) != 0 || len(node.Actors) != 0):
				priority = storage.MatchPriorityEventActor
			case !node.SourceMatch.ByActor && (len(node.Roles) != 0):
				priority = storage.MatchPriorityEventRole
			case !node.SourceMatch.ByActor:
				priority = storage.MatchPriorityEvent
			case node.SourceMatch.ByActor && (len(node.ActorIDs) != 0 || len(node.Actors) != 0):
				priority = storage.MatchPriorityActor
			case node.SourceMatch.ByActor && (len(node.Roles) != 0):
				priority = storage.MatchPriorityRole
			default:
				panic("could not determine priority")
			}

			return &Match{
				Pipeline: p,
				Output: storage.NodeOutputRef{
					NodeID: node.ID,
					Output: output,
				},
				Priority: priority,
			}
		}
	}
	return nil
}

// GetNode returns a node from its ID
func (p Pipeline) GetNode(id string) (storage.Node, error) {
	for _, node := range p.Graph.Nodes {
		if node.ID == id {
			return node, nil
		}
	}
	return storage.Node{}, fmt.Errorf("No such node: %s", id)
}
